<?php

namespace HelpersLibrary;

// Prevent Direct Access
( defined( 'ABSPATH' ) ) || die;

/**
 * Sanitizes all $_REQUEST, $_GET, and $_POST variables reliably
 */
final class REQUEST {

	/**
	 * Retrieves and sanitizes the supplied key as a boolean.
	 *
	 * @param string $key
	 * @param string $subkey
	 * @param bool   $default
	 *
	 * @return bool
	 */
	public static function bool( string $key = '', string $subkey = '', bool $default = false ): bool {

		$value = $default;

		if ( $subkey && isset( $_REQUEST[ $key ][ $subkey ] ) ) {
			$value = (bool) $_REQUEST[ $key ][ $subkey ];
		} elseif ( isset( $_REQUEST[ $key ] ) ) {
			$value = (bool) $_REQUEST[ $key ];
		}

		return $value;
	}

	/**
	 * Retrieves and sanitizes the supplied key as a text field.
	 *
	 * @param string $key
	 * @param string $subkey
	 * @param string $default
	 *
	 * @return string
	 */
	public static function text_field( string $key = '', string $subkey = '', string $default = '' ): string {

		$value = $default;

		if ( $subkey && isset( $_REQUEST[ $key ][ $subkey ] ) ) {
			$value = sanitize_text_field( $_REQUEST[ $key ][ $subkey ] );
		} elseif ( isset( $_REQUEST[ $key ] ) ) {
			$value = sanitize_text_field( $_REQUEST[ $key ] );
		}

		return $value;
	}

	/**
	 * Retrieves and sanitizes the supplied key as an integer.
	 *
	 * @param string $key
	 * @param string $subkey
	 * @param int    $default
	 *
	 * @return int
	 */
	public static function int( string $key = '', string $subkey = '', int $default = 0 ): int {

		$value = $default;

		if ( $subkey && isset( $_REQUEST[ $key ][ $subkey ] ) ) {
			$value = (int) $_REQUEST[ $key ][ $subkey ];
		} elseif ( isset( $_REQUEST[ $key ] ) ) {
			$value = (int) $_REQUEST[ $key ];
		}

		return $value;

	}

	/**
	 * Retrieves and sanitizes the supplied key as a key string.
	 *
	 * @param string $key
	 * @param string $subkey
	 * @param string $default
	 * @param string $prepend
	 * @param string $append
	 *
	 * @return string
	 */
	public static function key( string $key = '', string $subkey = '', string $default = '', string $prepend = '', string $append = '' ): string {

		$value = $default;

		if ( $subkey && isset( $_REQUEST[ $key ][ $subkey ] ) ) {
			$value = sanitize_key( $_REQUEST[ $key ][ $subkey ] );
		} elseif ( isset( $_REQUEST[ $key ] ) ) {
			$value = sanitize_key( $_REQUEST[ $key ] );
		}

		return $prepend . $value . $append;

	}

	/**
	 * Retrieves and sanitizes the supplied key's array values
	 *
	 * @param string $key
	 * @param array  $default
	 * @param string $sanitize
	 *
	 * @return array
	 */
	public static function array( string $key = '', array $default = [], string $sanitize = 'text_field' ): array {

		$array = [];

		if ( isset( $_REQUEST[ $key ] ) ) {

			if ( is_array( $_REQUEST[ $key ] ) ) {

				// Sanitize the array values
				if ( ! empty( $_REQUEST[ $key ] ) ) {

					if ( 'text_field' === $sanitize ) {
						$array = map_deep( $_REQUEST[ $key ], 'sanitize_text_field' );
					} elseif ( 'key' === $sanitize ) {
						$array = map_deep( $_REQUEST[ $key ], 'sanitize_key' );
					} elseif ( 'int' === $sanitize ) {
						$array = map_deep( $_REQUEST[ $key ], 'intval' );
					} elseif ( 'bool' === $sanitize ) {
						$array = map_deep( $_REQUEST[ $key ], 'boolval' );
					}

				}

			} else {

				// Sanitize and create an array
				if ( 'text_field' === $sanitize ) {
					$array[] = sanitize_text_field( $_REQUEST[ $key ] );
				} elseif ( 'key' === $sanitize ) {
					$array[] = sanitize_key( $_REQUEST[ $key ] );
				} elseif ( 'int' === $sanitize ) {
					$array[] = (int) $_REQUEST[ $key ];
				} elseif ( 'bool' === $sanitize ) {
					$array[] = (bool) $_REQUEST[ $key ];
				}

			}

		} else {
			// Fallback on the default value
			$array = $default;

			// Set the original with the default value to make compatible with saving empty settings values
			$_REQUEST[ $key ] = $default;
		}

		return $array;

	}

	/**
	 * Retrieves and sanitizes the entire $_GET array
	 *
	 * @param array $empty
	 *
	 * @return array
	 */
	public static function GET( array $empty = [] ): array {

		return self::var( 'GET', $empty );

	}

	/**
	 * Retrieves and sanitizes the entire $_POST array
	 *
	 * @param array $empty
	 *
	 * @return array
	 */
	public static function POST( array $empty = [] ): array {

		return self::var( 'POST', $empty );

	}

	/**
	 * Retrieves and sanitizes the entire $_REQUEST array
	 *
	 * @param array $empty
	 *
	 * @return array
	 */
	public static function REQUEST( array $empty = [] ): array {

		return self::var( 'REQUEST', $empty );

	}

	/**
	 * Retrieves and sanitizes the entire $_SERVER array
	 *
	 * @param array $empty
	 *
	 * @return array
	 */
	public static function SERVER( array $empty = [] ): array {

		return self::var( 'SERVER', $empty );

	}

	/**
	 * Retrieves and sanitizes the supplied $_SERVER key as a text field.
	 *
	 * @param string $key
	 * @param string $subkey
	 * @param string $default
	 *
	 * @return string
	 */
	public static function SERVER_text_field( string $key = '', string $subkey = '', string $default = '' ): string {

		$value = $default;

		if ( $subkey && isset( $_SERVER[ $key ][ $subkey ] ) ) {
			$value = sanitize_text_field( $_SERVER[ $key ][ $subkey ] );
		} elseif ( isset( $_SERVER[ $key ] ) ) {
			$value = sanitize_text_field( $_SERVER[ $key ] );
		}

		return $value;
	}

	/**
	 * Retrieves and sanitizes $_GET, $_POST, and $_REQUEST arrays
	 *
	 * @param string $type
	 * @param array  $empty
	 *
	 * @return array
	 */
	private static function var( string $type = 'GET', array $empty = [] ): array {

		$array = [];

		if ( 'GET' === $type && isset( $_GET ) && ! empty( $_GET ) ) {
			$array = map_deep( $_GET, 'sanitize_text_field' );
		} elseif ( 'POST' === $type && isset( $_POST ) && ! empty( $_POST ) ) {
			$array = map_deep( $_POST, 'sanitize_text_field' );
		} elseif ( 'REQUEST' === $type && isset( $_REQUEST ) && ! empty( $_REQUEST ) ) {
			$array = map_deep( $_REQUEST, 'sanitize_text_field' );
		} elseif ( 'SERVER' === $type && isset( $_SERVER ) && ! empty( $_SERVER ) ) {
			$array = map_deep( $_REQUEST, 'sanitize_text_field' );
		}

		// To ensure blank settings are set
		if ( ! empty( $empty ) ) {
			foreach ( $empty as $key => $type ) {
				if ( ! isset( $array[ $key ] ) ) {
					if ( 'array' === $type ) {
						$array[ $key ] = [];
					} elseif ( 'string' === $type ) {
						$array[ $key ] = '';
					} elseif ( 'int' === $type ) {
						$array[ $key ] = 0;
					}
				}
			}
		}

		return $array;

	}

}